#! /usr/bin/env python

#
# Copyright(C) 2005,2006,2007 INL
# Written by Jean Gillaux <jean@inl.fr>
#            Damien Boucard <damien.boucard@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


__doc__ = """

This module is responsible for parsing xml nufw firewall desccription, and mapping it into objects.

"""
import xml.dom.minidom
import IPy
from os import linesep
from net.defs import INTERNET

ZONE_EXT = "external"
ZONE_INT = "internal"
ZONE_VPN = "vpn"

DFT_GATEWAY = "dftgateway"

class UnknownVpn(Exception):
    def __init__(self, mark):
        self.mark = mark

    def __str__(self):
        return 'unknown vpn with mark %s'%(self.mark)

class Net:
    """Maps <net> tag from a network description xml file (nuface interface)
    """
    def __init__(self, name, addr, zone = '', type = '', mark = '', remote = '', id = None, enabled = 1):
        if(zone==''):
            self.__zone = ZONE_INT
        else:
            self.__zone = zone
        self.__conns = []
        if not addr: addr = '0.0.0.0/0'
        self.__addr = IPy.IP(addr)
        self.__name = name
        self.__type = type
        self.__mark = mark
        self.__remote = remote
        self.local_id = None
        self.__id = 0
        self.__enabled = enabled
        if id is not None: self.__id = int(id)

    def add_conn(self, dir, fwid, iface, addr = None, gw = None,
            dftgw = None, key=None, side=None):
        h = {'direct':dir, 'fwid':fwid, 'iface':iface, 'addr': addr,
                'key': key, 'side': side}
        #if(len(args)==1):
        #    h['gateway'] = args[0]
        if gw:
            h['gateway'] = gw
        if dftgw:
            h[DFT_GATEWAY] = dftgw
        self.__conns.append(h)

    def get_conn(self, fwid):
        for conn in self.__conns:
            if conn['fwid']==fwid:
                return conn

    def iter_conns(self):
        for index in range(0, len(self.__conns)):
            yield self.__conns[index]

    def get_conns(self):
        """get connexions"""
        return self.__conns

    #getters
    def netmask(self):
        """get the netmask of the network
        """
        return self.__addr.netmask()

    def ip(self):
        """return the string representing the addr of the Net object"""
        return str(self.__addr)

    def addr(self):
        return self.__addr

    def name(self):
        return self.__name

    def zone(self):
        return self.__zone

    def type(self):
        return self.__type

    def mark(self):
        return self.__mark

    def remote(self):
        return self.__remote


    def is_vpn(self):
        """Checks if the network is a vpn """
        return self.__mark != '' or self.local_id != ''

    def is_enabled(self):
        return self.__enabled == "1"

    def id(self):
        """Get network id"""
        return self.__id

    def __str__(self):
        s = 'name: %s, addr:%s, id:%s, zone:%s' %(self.__name, str(self.__addr), str(self.__id), self.__zone)
        return s
    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if self.__id == other.__id and self.__addr == other.__addr:
            return True
        return False
    def __ne_(self, other):
        return not self.__eq__(other)
#End class Net

class Interface:
    """An interface has an ip addresse and a list of networks attached to it.
    """
    def __init__(self, id, name):
        self.__id = id
        self.__directnets = []
        self.__addr = {}
        self.__name = name
        self.__distnets = []
        self.__vpns = []
        self.__nets = []
        self.__gateways = {}
        self.__addr2net = {}
        self.__distnet2addr = {}

    def name(self) :
        return self.__name

    def add_addr(self, id, addr):
        assert id not in self.__addr
        self.__addr[id] = IPy.IP(addr)

    def add_directnet(self, net, vpn=False):
        if vpn:
            self.__vpns.append(net)
        else:
            addr_in_net = False
            for addr in self.__addr.values():
                if addr in net.addr():
                    addr_in_net = True
                    self.__addr2net[addr] = net
            if not addr_in_net:
                raise ValueError("Network %s (%s) doesn't match to any firewall" % (net.name(), net.addr()))

        self.__directnets.append(net)

    def add_distnet(self, net, gateway):
        self.__distnets.append(net)
        gateway_net = None
        related_addr = None
        for addr, direct in self.__addr2net.iteritems():
            direct = direct.addr()
            if gateway in direct and (gateway_net is None or direct < gateway_net):
                gateway_net = direct
                related_addr = addr
        assert related_addr is not None, "Interface %s cannot directly reach %s, the gateway of %s" %(self.__name, gateway, net)
        self.__distnet2addr[net.addr()] = related_addr

    def add_net(self, net):
        self.__nets.append(net)
            #if

    def distnets(self):
        for n in self.__distnets:
            yield n

    def vpns(self):
        for n in self.__vpns:
            yield n

    def allnets(self):
        """return all the network of the interface
        """
        for n in self.__nets:
            yield n

##    def netmask(self):
##        if(type(self.__net)==type(None)):
##            return ''
##        return self.__net.netmask()

    def addr(self):
        return map(str, self.__addr.values())
    def addr2(self):
        return self.__addr.values()
    def get_addr_by_id(self, id):
        return str(self.__addr[id])

    def id(self):
        return self.__id

    def directnets(self):
        return self.__directnets[:]

    def self_net(self):
        """return a Net object that represents the ip addr of the interface"""
        return Net(self.__name, None)
        pass

    def get_net_by_addr(self, addr):
        net =  self.__addr2net.get(addr)
        if net is None:
            raise KeyError('%s not found in %s' %(addr, self.__addr2net))
        return net
    
    def get_local_ip_connected_to_distnet(self, distnet):
        assert(distnet in self.__distnets)
        return self.__distnet2addr[distnet.addr()]
    
    def deb_conf(self):
        keys = self.__addr.keys()
        keys.sort()
        addr = self.__addr[keys[0]]
        s = """
auto %(name)s
iface %(name)s inet static
    address %(address)s
    netmask %(netmask)s""" %{'name':self.__name,
                             'address':addr,
                             'netmask':self.__addr2net[addr].netmask()
                            }

        cpt = 0
        for key in keys[1:]:
            addr = self.__addr[key]
            s += """
auto %(name)s:%(ext)s
iface %(name)s:%(ext)s inet static
    address %(address)s
    netmask %(netmask)s""" %{'name':self.__name,
                             'ext':cpt,
                             'address':addr,
                             'netmask':self.__addr2net[addr].netmask()
                            }
            cpt += 1
        return s

    def __str__(self):
        return "<iface id=%s addr=%s>" %(self.__id, self.__addr)

#End class Interface
class RuleList(object):
    """List of rules
    """
    def __init__(self):
        self.rules = {}
        self.rule_list = []
        self.bad = 0

    def is_bad(self):
        return self.bad == 1

    def add_rule(self, rule, order):
        self.rule_list.append((order, rule))
##        rule_list = self.rules.get(order)
##        if rule_list is None:
##            self.rules[order] = [rule]
##        else:
##            self.rules[order].append(rule)

    def ordered_rules(self):
##        keys = self.rules.keys()
##        keys.sort()
##        result = []
##        for id in keys:
##            result.extend([(id, rule) for rule in self.rules[id]])
        self.rule_list.sort()
        return self.rule_list[:]
##        assert result == self.rule_list
##        return result


class FWNetConf(object):
    """Dns configuration for firewall
    """
    def __init__(self):
        self.__dns_search = []
        self.__dns_nameservers = []
        pass

    def add_dns_search(self, resolv):
        """add a search item for dns configuration.
        parameters: resolv = str
        """
        self.__dns_search.append(resolv)

    def add_dns_nameserver(self, server):
        """Add a dns namserver.
        parameters: server = str. the ip addr of the namserver
        """
        self.__dns_nameservers.append(server)

    def get_dns_search(self):
        """Return list of dns resolv domains
        """
        return self.__dns_search
    def get_dns_nameservers(self):
        """Return the list of dns nameservers
        """
        return self.__dns_nameservers

    def get_dns_conf(self):
        """Get dns configuration. Returns (resolv list, nameservers list)
        """
        return (self.__dns_search, self.__dns_nameservers)
#End class FWNetConf

class FW:
    """FW object is a firewall with:
- Interfaces
- Nets
- Rules
Rules are stored in chains corresponding to their source/destination ip addresses.
    """
    def __init__(self, id):
        self.__ifaces = []
        self.__nets = []
        self.__id = id
        self.__hrules = {}
        self.__allrules = []
        self.__nat_rules = {'snat': [], 'dnat': [], 'pnat': []}
        self.__chains = []
        self.__ext_net = None
        self.__vpns = {}
        self.manage_input = True
        self.manage_output = True
        self.manage_forward = True
        self.__netconf = None
        self.queue = None
        self.auth_ext = False #authenticate packets from internet?
        self.no_same_iface = False #don't generate rules with same input and output interface?
        self.no_check_net = False # don't check inclusion and overlaping on enterprise networks
        self.__same_counter=0 #for debug only


    def id(self):
        """get firewall id"""
        return self.__id

    def has_dft_net(self):
        return self.__ext_net is not None

    def add_iface(self, iface):
        self.__ifaces.append(iface)

    def ifaces(self):
        for i in self.__ifaces:
            yield i

    def add_net(self, net, conn = None):
        self.__nets.append(net)
        #self.__chains = self.chains(all = 1)
        #if net.zone()==ZONE_EXT: #TODO: test & remove
        #    self.__ext_net = net
        #More logical than net.zone():
        if conn is not None and conn.has_key(DFT_GATEWAY):
            self.__ext_net = net
        if net.is_vpn():
            self.__vpns[net.id()] = net


    #  Checks if a network is the default route network for the firewall
    def is_default_net(self, net):
        """Checks if a network is the default route network for the firewall
           net: Net object, the network to be checked
        """
        if not self.__ext_net:
            return False
        return net.id() == self.__ext_net.id()

    def is_normal_net(self, net):
        """Check if a network is not default route nor vpn"""
        return (net.is_vpn() is not True and self.is_default_net(net) is not True)

    def deb_conf(self):
        """Returns a debian /etc/interfaces network configuration
        """
        s = ''
        for myiface in self.__ifaces:
            s+="""%s
    %s""" %(myiface.deb_conf(), self.debconf_gateways(myiface))
        return s

    def debconf_gateways(self, iface):
        """get 'route add' commands for network behind interface, for inclusion in /etc/network/interface post-up sections
        """
        s = ''
        nets = iface.directnets()
        for n in nets:
            h = n.get_conn(self.__id)
            if h.has_key(DFT_GATEWAY):
                s += 'gateway %s%s'%(h[DFT_GATEWAY], linesep)
        for n in iface.distnets():
            h = n.get_conn(self.__id)
            #if h.has_key('gateway') and n.zone() not in [ZONE_VPN]: #do not use zones anymore
            if h.has_key(DFT_GATEWAY):
                    s += 'gateway %s%s'%(h[DFT_GATEWAY], linesep)
            if h.has_key('gateway') and not n.is_vpn():
                s += 'post-up route add -net %s gw %s%s'%(n.ip(), h['gateway'], linesep)
        return s

    def routes(self):
        s = ''
        for iface in self.__ifaces:
            for n in iface.distnets():
                h = n.get_conn(self.__id)
                s+="%s\t%s\n" %(str(n.addr().net()), h['router'])
        return s

    def __add_chain_for_rules_if_not_exist(self, chain):
        if not self.__hrules.has_key(chain):
            self.__hrules[chain] = RuleList()

    def find_zone(self, ip, mark):
        """Find the zone of the firewall which contains the string ip address "ip" and/or
        the string mark
        """
        if mark != '' and self.__vpns.has_key(mark):
            return ZONE_VPN
        for net in self.__nets:
            if net.zone != ZONE_EXT and ip in net.addr():
                return net.zone()
        return ZONE_EXT

    def is_correct_network(self, network):
        """Check if network is correct
        network should be included in one enterprise network
        if not: network must not overlap with enterprise network
        """
        net = IPy.IP(network)
        l = [IPy.IP(n) for n in self.entreprise_nets()]
        included = 0
        overlap = 0
        for net2 in l:
            if net in net2:
                included = 1
        if included:
            return True
        else:
            for nete in l:
                if nete.overlaps(net):
                    return False
            return True

    def get_local_ip_connected_to(self, remote_ip):
        """Returns ip address of fw which can route to remote_ip
        """
        remote = IPy.IP(remote_ip)
        internet = IPy.IP('0.0.0.0/0')
        to_internet = (remote == internet)
        best_candidate = None
        candidate_dict = None
        dftgw_dict = None
        for iface in self.__ifaces:
            for addr in iface.addr2():
                net = iface.get_net_by_addr(addr)
                if not to_internet and remote in net.addr() and (best_candidate is None or net.addr() < best_candidate):
                    best_candidate = net.addr()
                    candidate_dict = {'iface':iface, 'addr':addr}
                    continue
                if net.get_conn(self.__id).has_key(DFT_GATEWAY):
                    dftgw_dict = {'iface':iface, 'addr':addr}
                    if to_internet:
                        return dftgw_dict
            for net in iface.distnets():
                conn = net.get_conn(self.__id)
                if not to_internet and remote in net.addr() and (best_candidate is None or net.addr() < best_candidate):
                    addr = iface.get_local_ip_connected_to_distnet(net)
                    best_candidate = net.addr()
                    candidate_dict = {'iface':iface, 'addr':addr}
                    continue
                assert(not (to_internet and conn.has_key(DFT_GATEWAY)))
            for net in iface.vpns():
                conn = net.get_conn(self.__id)
                if not to_internet and remote in net.addr() and (best_candidate is None or net.addr() < best_candidate):
                    vpn_dict = self.get_local_ip_connected_to(net.remote())
                    assert vpn_dict['iface'] == iface, "Interface %s cannot reach %s, the remote machine routing the VPN %s" %(iface.name(), net.remote(), net)
                    best_candidate = net.addr()
                    candidate_dict = vpn_dict
                assert(not (to_internet and conn.has_key(DFT_GATEWAY)))
        if candidate_dict is None:
            assert dftgw_dict is not None, "No route to %s" %(remote_ip)
            candidate_dict = dftgw_dict
        return candidate_dict

    def chain(self, ip_src, ip_dst, zone_src = '', zone_dst = '', mark_src = '', mark_dst = ''):
        """Returns the fw chain that fits ip_src/ip_dst/zone_src/zone_dst/mark
        """
        # FIXME / TODO
        # do not use anymore zone attribute for elements. (OK for networks)
        chp1 = None
        chp2 = None
        if1 = None
        if2 = None
        find1 = 0
        find2 = 0
        ip_src = IPy.IP(ip_src)
        ip_dst = IPy.IP(ip_dst)
        #Let's find the right network. no need of rules zones.
        for iface in self.__ifaces:
            net_id_by_broadcast = dict([ (net.addr().broadcast(), net.id()) for net in iface.directnets() ])
            iface_addr = iface.addr2()
            for net in iface.allnets():
                #if not self.is_default_net(net): #we don't want to match default net for now: it comes at last
                # got source network?
                if ip_src in net.addr():
                    # we want the smallest network
                    if chp1 is not None and net.addr() <= chp1.addr():
                        chp1 = net
                        find1 = 1
                    elif chp1 is None:
                        chp1 = net
                        find1 = 1
                # got destination network?
                if ip_dst in net.addr():
                    # we want the smallest network
                    if chp2 is not None and net.addr() <= chp2.addr():
                        chp2 = net
                        find2 = 1
                    elif chp2 is None:
                        chp2 = net
                        find2 = 1
            #INPUT or OUTPUT: all interfaces, default included
            #create fake network for input/output
            iface_net_id = None
            if ip_src in iface_addr :
                iface_net_id = iface.get_net_by_addr(ip_src).id()
            else:
                iface_net_id = net_id_by_broadcast.get(ip_src)
            if iface_net_id is not None:
                chp1 = Net('IF', ip_src.strNormal(), type = 'ipv4', zone = ZONE_INT, id=(-1))
                chp1.add_conn('0', self.__id, iface.id())
                if1 = iface
                find1 = 1

            iface_net_id = None
            if ip_dst in iface_addr :
                iface_net_id = iface.get_net_by_addr(ip_dst).id()
            else:
                iface_net_id = net_id_by_broadcast.get(ip_dst)
            if iface_net_id is not None:
                chp2 = Net('IF', ip_dst.strNormal(), type = 'ipv4', zone = ZONE_INT, id=(-1))
                chp2.add_conn('0', self.__id, iface.id())
                if2 = iface
                find2 = 1
        #have we find what we want? if not, got to the evil default route.
        #FIXME NOINT what todo if no ext_net?
        if not chp1:
            chp1 = self.__ext_net
            find1 = 1
        if not chp2:
            chp2 = self.__ext_net
            find2 = 1
        if not if1:
            if1 = self.get_iface_by_net(chp1)
        if not if2:
            if2 = self.get_iface_by_net(chp2)
        return {'chain':chp1.name()+'-'+chp2.name(), 'netsrc':chp1, 'netdst':chp2, 'ifsrc': if1, 'ifdst': if2}

    def chains(self, all = 0, inout = 0):#TODO: get INPUT:OUTPUT chains with inout = 1
        """returns a list of firewall's chains.
all = 1 : all networks are considered , including distant ones
inout = 1: chains for input/output are also generated
        """
        l = []
        l2 = []
        for iface in self.__ifaces:
            #if((all==0 and iface.net().zone()==ZONE_INT) or all==1):
            for net in iface.directnets():
                if( (all==0 and self.is_normal_net(net)) or all==1):
                    l.append({'name':net.name(), 'net':net, 'zone':None, 'iface':iface})
            for n in iface.distnets():
                if(all == 1):
                    l.append({'name':n.name(), 'net':n, 'zone':None, 'iface':iface})
            if inout == 1:
                #create chain for iface IP and broadcast IP
                    for n in iface.allnets():
                        choutput = {'name':'IF-'+n.name(), 'net_dst':n, 'type':"output", 'if_src':iface, 'if_dst':iface}
                        self.__add_chain_for_rules_if_not_exist('IF-'+n.name())
                        chinput = {'name':n.name()+'-IF', 'net_src':n, 'type':"input", 'if_src':iface, 'if_dst':iface}
                        self.__add_chain_for_rules_if_not_exist('%s-IF'%(n.name()))
                        l2.extend([choutput, chinput])

                #l.append({'name':iface.name(), 'net':iface.self_net(), 'zone':iface.net().zone(), 'iface':iface})
        def _fun(l1, l2):
            res = []
            for u1 in l1:
                for u2 in l2:
                    if(u1['name']!=u2['name']):
                        if self.no_same_iface and u1['iface'] == u2['iface']:
                            self.__same_counter += 1
                        else:
                            res.append({'name':u1['name']+'-'+u2['name'], 'net_src':u1['net'], 'net_dst':u2['net'], 'type':"forward" ,'if_src':u1['iface'], 'if_dst':u2['iface']})
                            if not self.__hrules.has_key(u1['name']+'-'+u2['name']):
                                self.__hrules[u1['name']+'-'+u2['name']] = RuleList()
            return res
        tmpres = _fun(l, l)
        tmpres.extend(l2)
        return tmpres


    def entreprise_nets(self):
        l = []
        for iface in self.__ifaces:
            for n in iface.allnets():
                #if(n.zone()==ZONE_INT): #no more zone
                if self.is_normal_net(n):
                    l.append(str(n.addr()))
        return l

    def add_rule(self, r):
        """Add a rule to the firewall into the right chain
        """
        #if r.is_auto == 1: print 'rule is auto',r
        elt_src = r.elt_src()
        elt_dst = r.elt_dest()#find and set the zones if they're not set
        assert(elt_src.get('numnet') or elt_dst.get('numnet'))
        if elt_src.get('numnet'):
            netsrc = IPy.IP(elt_src.get('numnet'))
        else:
            netsrc = self.get_local_ip_connected_to(elt_dst.get('numnet'))['addr']
        if elt_dst.get('numnet'):
            netdst = IPy.IP(elt_dst.get('numnet'))
        else:
            netdst = self.get_local_ip_connected_to(elt_src.get('numnet'))['addr']
        # check inclusion and overlaping for dest/src networks with enterprise networks
        if not self.no_check_net:
            if netsrc != INTERNET and not self.is_correct_network(str(netsrc)):
                raise ValueError("network not correct: %s"%(elt_src.get('numnet')))
            if netdst != INTERNET and not self.is_correct_network(str(netdst)):
                raise ValueError("network not correct: %s"%(elt_dst.get('numnet')))
        ch = self.chain(str(netsrc), str(netdst), zone_src = r.elt_src().get('zone'), zone_dst = r.elt_dest().get('zone'), mark_src = r.elt_src().get('mark'), mark_dst = r.elt_dest().get('mark'))
        #cpl used to get rule order in chain
        chain_cpl_id = (ch['netsrc'].id(), ch['netdst'].id())
        r_order = None
        if r.sort_order.has_key(chain_cpl_id):
            r_order = r.sort_order[chain_cpl_id]
        r.set_netsrc(ch['netsrc'])
        r.set_netdst(ch['netdst'])
        r.snetsrc = str(ch['netsrc'].addr())
        r.snetdst = str(ch['netdst'].addr())
        if(not self.__hrules.has_key(ch['chain'])):
            self.__hrules[ch['chain']] = RuleList()
            self.__hrules[ch['chain']].bad = 1
        # add rule only if it has an order in the bichain
        if r_order is not None:
            self.__hrules[ch['chain']].add_rule(r, r_order)
            self.__allrules.append(r)
        else:
            raise ValueError("acl %s: no descsort in %s"%(r.aclid, str(chain_cpl_id)))

    def set_nat_rules(self, snat, dnat, pnat):
        """Set NAT rules for FW object
        """
        # set rules interfaces
        if self.__ext_net:
            ext_iface = self.get_iface_by_net(self.__ext_net).name()
        for r in snat:
            d = self.chain(r._to.strNormal(), r._to.strNormal())
            r._iface = d['ifdst'].name()
        # nat/pnat: see if destination address is set
        for r in dnat:
            if r._src:
                d = self.chain(r._src.strNormal(), r._src.strNormal())
                r._iface = d['ifdst'].name()
            else:
                r._iface = ext_iface
        for r in pnat:
            if r._dst:
                d = self.chain(r._dst.strNormal(), r._dst.strNormal())
                r._iface = d['ifdst'].name()
        self.__nat_rules = {'snat': snat, 'dnat': dnat, 'pnat': pnat}

    def find_network(self, addr):
        """Given a ip address, find the network that this ip belongs to"""
        best_net = None
        for iface in self.__ifaces:
            for net in iface.allnets():
                if not self.is_default_net(net):
                    if (best_net is None or net.addr() in best_net.addr()) and addr in net.addr():
                        best_net = net
        if best_net is not None:
            return best_net
        return self.__ext_net#FIXME NOINT

    def rule_dst_is_src(self, r):
        """Test if a rule comes from and goes to the same network """
        src = r.elt_src(); net_src = None
        dst = r.elt_dest(); net_dst = None
        #if src.get('mark') != '': net_src = self.__vpns[src.get('mark')]
        #if dst.get('mark') != '': net_dst = self.__vpns[src.get('mark')]
        if net_src is None:
            net_src = self.find_network(src.get('numnet'))
        if net_dst is None:
            net_dst = self.find_network(dst.get('numnet'))
        if self.is_default_net(net_src) and self.is_default_net(net_dst): return True
        if net_dst.addr() in net_src.addr() and net_src.addr() in net_dst.addr(): return True
        return False


    def rule_inout(self, rule):
        """Test if the rule has its src/dest = one of the fw ifaces"""
        return self.rule_input(rule) or self.rule_output(rule)

    def rule_input(self, rule):
        """Test if the rule has its dest = one of the fw ifaces"""
        if rule.elt_dest().get('is_local'):
            # Easy case : elt_dest is a local_ipv4
            return True
        iptest = rule.elt_dest().get('numnet')

        iptest = IPy.IP(iptest)
        for iface in self.__ifaces:
            if iptest in [ net.addr().broadcast() for net in iface.directnets() ] :
                return True
            for ifaceaddr in iface.addr():
                iface_addr = IPy.IP(ifaceaddr)
                if iface_addr == iptest: return True
        return False

    def rule_output(self, rule):
        """Test if the rule has its source = one of the fw ifaces or network broadcast"""
        if rule.elt_src().get('is_local'):
            # Easy case : elt_src is a local_ipv4
            return True
        iptest = rule.elt_src().get('numnet')
        iptest = IPy.IP(iptest)
        for iface in self.__ifaces:
            if iptest in [ net.addr().broadcast() for net in iface.directnets() ] :
                return True
            for ifaceaddr in iface.addr():
                iface_addr = IPy.IP(ifaceaddr)
                if iface_addr == iptest: return True
        return False

    def chain_input(self, chain):
        """Test if the chain has its dest = one of the fw iface address or network broadcast"""
        return chain['type'] == "input"

    def chain_output(self, chain):
        """test if the chain has its source = one of the fw iface address or network broadcast"""
        return chain['type'] == "output"

    def chain_inout(self, chain):
        """test if the chain has its src or dst = one of the fw ifaces address or network broadcast"""
        return self.chain_input(chain) or self.chain_output(chain)

    def is_iface_addr_broadcast(self, addr):
        """Test if the string addr is one of the fw ifaces addresses or iface broadcast addr"""
        for iface in self.__ifaces:
            if addr in [ net.addr().broadcast().strNormal() for net in iface.directnets() ]:
                return True
            for ifaceaddr in iface.addr():
                if ifaceaddr == addr: return True
        return False

    def hrules(self):
        return self.__hrules

    def allrules(self):
        return self.__allrules

    def get_nat_rules(self):
        return [self.__nat_rules['snat'], self.__nat_rules['dnat'], self.__nat_rules['pnat']]

    def set_netconf(self, net_conf):
        """set netconf member of FW object
        """
        self.__netconf = net_conf

    def get_netconf(self):
        """Get Net Conf of Firewall
        returns: FwNetConf object
        """
        return self.__netconf

    def get_iface_by_net(self, net):
        """ Get iface corresponding to a specific network"""
        for tmpiface in self.__ifaces:
            if net in tmpiface.directnets():
                return tmpiface
        return None

    def get_iface_by_id(self, id):
        """get firewall iface referenced by its id in xml"""
        for iface in self.__ifaces:
            if int(iface.id()) == int(id):
                return iface
        return None

    def get_net_by_id(self, id):
        """Find a network identified by its ID"""
        id = int(id)  # In case the parameter is a string or a unicode.
        for net in self.__nets:
            if net.id() == id:
                return net
        return None

    def get_ext_net(self):
        """get default network of firewall"""
        return self.__ext_net

    def get_vpns(self):
        """Returns vpn networks of firewall"""
        return self.__vpns

    def get_local_ips(self):
        """Returns ip adresses of firewall"""
        ips = []
        for iface in self.__ifaces:
            ips.extend(map(str, iface.addr()))
        return ips
#End class FW


def _xml_parse_fwnetconf(xmlnode):
    """returns a FWNetConf object
    parameters: xmlnode = XmlNode (fw node)
    """
    ret = FWNetConf()
    dns = xmlnode.getElementsByTagName('dns')
    if not dns: return None
    for mynode in dns[0].childNodes:
        if mynode.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and mynode.nodeName == 'search':
            if mynode.hasAttribute('val'):
                ret.add_dns_search(mynode.getAttribute('val'))
        elif mynode.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and mynode.nodeName == 'nameserver':
            if mynode.hasAttribute('val'):
                ret.add_dns_nameserver(mynode.getAttribute('val'))
    return ret


def parse(file):
    """Parse a file that follows nuface firewall description xml format.
Returns a list of FW objects.
    """
    fwhash = {}
    doc = xml.dom.minidom.parse(file)
    fwl = []
    netl = {}

    # Parse networks
    eltnets = doc.getElementsByTagName('nets')
    eltsnet = eltnets[0].childNodes
    for eltnet in eltsnet:
        if(eltnet.nodeType==xml.dom.minidom.Node.ELEMENT_NODE and eltnet.nodeName=="net"):
            cnet = Net(eltnet.getAttribute('name'), eltnet.getAttribute('addr'), zone = None, type = eltnet.getAttribute('type'), mark = eltnet.getAttribute('mark'), remote = eltnet.getAttribute('remote'), id = eltnet.getAttribute('id'), enabled = eltnet.getAttribute('enabled'))
            cnet.local_id = eltnet.getAttribute('local_id')
            if(eltnet.hasChildNodes()):
                for conn in eltnet.getElementsByTagName('connection'):
                    fwid = conn.getAttribute('fwid')
                    direct = conn.getAttribute('direct')
                    iface = conn.getAttribute('iface')
                    addr = conn.getAttribute('addr')
                    key = conn.getAttribute('key')
                    side = conn.getAttribute('side')
                    #sd = ''
                    if(direct=='1'):
                        #sd = 'directe'
                        cnet.add_conn(direct, fwid, iface, dftgw = conn.getAttribute(DFT_GATEWAY))
                    else:
                        #sd = 'indirecte'
                        cnet.add_conn(direct, fwid, iface, addr, gw =
                                conn.getAttribute("gateway"), dftgw =
                                conn.getAttribute(DFT_GATEWAY), key=key,
                                side=side)
                #put new found net into global nets list
                    if(not netl.has_key(fwid)): netl[fwid] = [cnet]
                    else: netl[fwid].append(cnet)

    # Parse firewalls
    for elt in doc.getElementsByTagName("fw"):
        fwid = elt.getAttribute('id')
        eltsinterface = elt.getElementsByTagName('interface')
        myfw = FW(fwid)
        # set queue number (NFQUEUE)
        if elt.hasAttribute('queue'):
            queue = elt.getAttribute('queue')
            if queue.isdigit():
                myfw.queue = queue
        myfw.set_netconf(_xml_parse_fwnetconf(elt))#get net conf attributes
        for interface in eltsinterface:
                myiface = Interface(interface.getAttribute('id'), interface.getAttribute("name"))
                eltsaddr = interface.getElementsByTagName('address')
                for addr in eltsaddr:
                    myiface.add_addr(addr.getAttribute('id'), addr.getAttribute("addr"))
                myfw.add_iface(myiface)
                distnets = []
                for net in netl[fwid]:
                    for conn in net.iter_conns():
                        if (conn['fwid']==fwid and conn['iface']==myiface.id()):
                            if conn.get('gateway'):  # Routed network.
                                distnets.append((net, conn['gateway']))
                            else:  # Direct net or VPN.
                                myiface.add_directnet(net,
                                        vpn=conn.get('direct') == '0')
                            myiface.add_net(net)
                            break # TODO: we have the connection
                    myfw.add_net(net, conn = conn)
                for dnet, gateway in distnets:
                    myiface.add_distnet(dnet, gateway)
        fwhash[fwid] = myfw
    return fwhash


