#!/bin/bash
SCRIPT_NAME=`basename $0`
SCRIPT_DIR=`dirname $0`
TMP_DIR=/tmp/nupyf-tests
LDAP_BASE="ou=acls,dc=test,dc=lld"

function print_usage {
  echo ""
  echo "Usage:"
  echo "  $SCRIPT_NAME [-h] [-f] [-l] [-i <INPUT_DIR>] [-a <ACL_FILE>] <testcase> <testname>"
  echo ""
  echo "Options:"
  echo "  -h : Prints this help."
  echo "  -i <INPUT_DIR> : Directory where taking expects files to add (by default $TMP_DIR)."
  echo "  -a <ACL_FILE> : File containing the ACLs used for this test (if not given, use the already existing file if it exits)."
  echo "  -l : Also adds LDAP contents into expect directory."
  echo "  -f : Force test overwriting if the acl file or expect directory already exist."
  echo ""
  echo "Arguments:"
  echo "  <testcase> : a testcase is a subdirectory included in the $SCRIPT_DIR. This subdirectory must have been created manually and contain a desc.xml file."
  echo "  <testname> : the base of the name of the test to add. By example, 'foo' will give an 'acls_foo.xml' ACL file and a 'expect_foo/' subdirectory."
} 

INPUT_DIR=
ACL_FILE=
force=1
ldap_mode=1
while getopts "hfli:a:" flag
do
  case "$flag" in
    'h') print_usage; exit 0;;
    'f') force=0;;
    'l') ldap_mode=0;;
    'i') INPUT_DIR="$OPTARG";;
    'a') ACL_FILE="$OPTARG";;
    '?') echo "Bad option: $flag"; print_usage; exit 1;;
  esac
done
shift $(($OPTIND-1))

if [ $# != 2 ]
then
  echo "Wrong number of arguments, see usage..."
  print_usage
  exit 2
else
  TESTCASE="$1"
  TESTNAME="$2"
fi
OUTPUT_ACL="$SCRIPT_DIR/$TESTCASE/$TESTNAME/acls.xml"
OUTPUT_EXPECT="$SCRIPT_DIR/$TESTCASE/$TESTNAME/expects"

# Checking -i option
if [ "$INPUT_DIR" ]
then
  if [ ! -d "$INPUT_DIR" ]
  then
    echo "Bad -i option: must be a directory containing the results of a nupyf process."
    print_usage
    exit 3
  fi
else
  if [ ! -d "$TMP_DIR" ]
  then
    echo "-i option is not given, but $TMP_DIR does not exist."
    print_usage
    exit 3
  fi
  INPUT_DIR="$TMP_DIR"
fi

# Checking <testcase> argument
if [ ! -d "$SCRIPT_DIR/$TESTCASE" ]
then
  echo "Bad <testcase> argument: '$TESTCASE' ; must be a subdirectory of '$SCRIPT_DIR'."
  print_usage
  exit 3
fi
if [ ! -e "$SCRIPT_DIR/$TESTCASE/desc.xml" ]
then
  echo "The file '$SCRIPT_DIR/$TESTCASE/desc.xml' is expected to be a valid testcase."
  print_usage
  exit 3
fi
if [ ! -e "$SCRIPT_DIR/$TESTCASE/nupyf.cmd" ]
then
  echo "The file '$SCRIPT_DIR/$TESTCASE/nupyf.cmd' is expected to be a valid testcase."
  print_usage
  exit 3
fi

# Checking testname directory existence
if [ $force = 1 -a -d "$SCRIPT_DIR/$TESTCASE/$TESTNAME" -a "$ACL_FILE" ]
then
  echo "The test directory already exists: $TESTNAME; maybe should you use the -f option."
  print_usage
  exit 3
fi

# Checking -a option
if [ "$ACL_FILE" ]
then
  if [ ! -f "$ACL_FILE" ]
  then
    echo "-a option: file not found: $ACL_FILE."
    print_usage
    exit 3
  elif [ ! -r "$ACL_FILE" ]
  then
    echo "-a option: permission denied: $ACL_FILE."
    print_usage
    exit 3
  fi
else
  if [ ! -d "$SCRIPT_DIR/$TESTCASE/$TESTNAME" ]
  then
    echo "-a option not given: the directory '$SCRIPT_DIR/$TESTCASE/$TESTNAME' must exist."
    print_usage
    exit 3
  fi
  if [ ! -f "$OUTPUT_ACL" ]
  then
    echo "-a option not given: the file '$OUTPUT_ACL' must exist."
    print_usage
    exit 3
  fi
  ACL_FILE="$OUTPUT_ACL"
fi

# Checking expected result directory existence (must be the last check)
if [ -e "$OUTPUT_EXPECT" ]
then
  if [ $force = 1 ]
  then
    echo "The resulting expect_dir already exists: $OUTPUT_EXPECT; maybe should you use the -f option."
    print_usage
    exit 3
  fi
  if [ ! -d "$OUTPUT_EXPECT" ]
  then
    echo "$OUTPUT_EXPECT exists but is not a directory."
    print_usage
    exit 3
  fi
  rm -rf "$OUTPUT_EXPECT"/*
fi

# Copying ACL file
mkdir -p "$SCRIPT_DIR/$TESTCASE/$TESTNAME"
cp "$ACL_FILE" "$OUTPUT_ACL.new"
mv -f "$OUTPUT_ACL.new" "$OUTPUT_ACL"

# Copying expected result
mkdir -p "$OUTPUT_EXPECT"
cp -v "$INPUT_DIR"/* "$OUTPUT_EXPECT/"

if [ $ldap_mode = 0 ]
then
    slapcat -s "$LDAP_BASE" | "$SCRIPT_DIR/"test_ldap.py "$OUTPUT_EXPECT/"ldap.ldif import
fi

echo Done.
