#!/bin/bash
#set -x
SCRIPT_NAME=`basename $0`
SCRIPT_DIR=`dirname $0`
NUPYF=/usr/sbin/nupyf
OUTPUT_DIR=/tmp/nupyf-tests
IPTABLES=/sbin/iptables
LDAP_HOST=localhost
LDAP_USER="cn=admin,dc=test,dc=lld"
LDAP_PASS=secret
LDAP_BASE="ou=acls,dc=test,dc=lld"

EXIT_VALUE=0

function print_usage {
  echo ""
  echo "Usage:"
  echo "  $SCRIPT_NAME [-hldc] [<testcase> [<testcase> ...]]"
  echo ""
  echo "Options:"
  echo "  -h : Prints this help."
  echo "  -l : Tests LDAP database."
  echo "  -d : Does not delete output files after failed test at $OUTPUT_DIR."
  echo "  -c : Continues the test procedure even if a test fails."
  echo ""
  echo "Arguments:"
  echo "  <testcase> : a testcase is a subdirectory included in the $SCRIPT_DIR."
  echo ""
  echo "If no argument is given, takes all testcases."
}

function check_readable {
  FILE_TO_CHECK="$1"
  if [ ! -f "$FILE_TO_CHECK" ]
  then
    echo "$FILE_TO_CHECK: file not found."
    return
  elif [ ! -r "$FILE_TO_CHECK" ]
  then
    echo "$FILE_TO_CHECK: permission denied."
  fi
}

function check_directory {
  FILE_TO_CHECK="$1"
  if [ ! -d "$FILE_TO_CHECK" ]
  then
    echo "$FILE_TO_CHECK: no such directory."
    return
  fi
}

function clear_ldap {
  /etc/init.d/slapd stop > /dev/null
  rm -f /var/lib/ldap/*
  slapadd -l "$SCRIPT_DIR"/ldap_minimal.ldif
  chown -R openldap:openldap /var/lib/ldap
  /etc/init.d/slapd start > /dev/null
}

function test_acl {
  ACL_DIR=$1
  if [ -e "$OUTPUT_DIR" ]
  then
    if [ -d "$OUTPUT_DIR" ]
    then
      rm -rf "$OUTPUT_DIR"
    else
      echo "$OUTPUT_DIR: file already exists but is not a directory."
      exit 43
    fi
  fi
  mkdir -p "$OUTPUT_DIR"
  ACL_XML="$ACL_DIR/acls.xml"
  check_readable "$ACL_XML"
  DESC_XML="$TEST_DIR/desc.xml"
  LDAP_OPT=
  if [ $test_ldap = 0 ]
  then
    clear_ldap
    LDAP_OPT="-s $LDAP_HOST -u \"$LDAP_USER\" -p \"$LDAP_PASS\" -b \"$LDAP_BASE\""
  fi
  echo "* Testing $ACL_DIR..."
  source $TEST_DIR/nupyf.cmd
  failure=1
  EXPECT_DIR="$ACL_DIR/expects"
  check_directory "$EXPECT_DIR"
  for OUTPUT_FILE in `ls $OUTPUT_DIR/*`
  do
    tail -n +2 $OUTPUT_FILE > $OUTPUT_FILE.tail
    mv -f $OUTPUT_FILE.tail $OUTPUT_FILE
    diff -u $EXPECT_DIR/`basename $OUTPUT_FILE` $OUTPUT_FILE
    diff_code=$?
    if [ ! $diff_code = 0 ]
    then
      failure=0
      echo "DIFF IN $OUTPUT_FILE"
    fi
  done
  if [ ! $failure = 0 -o $delete_output = 0 ]
  then
    rm -rf "$OUTPUT_DIR"
  fi
  if [ $test_ldap = 0 ]
  then
    slapcat -s "$LDAP_BASE" | "$SCRIPT_DIR"/test_ldap.py "$EXPECT_DIR"/ldap.ldif
    diff_code=$?
    if [ ! $diff_code = 0 ]
    then
      failure=0
      echo "DIFF IN LDAP"
    fi
  fi
  if [ $failure = 0 ]
  then
    echo "[FAIL]"
    if [ ! $continue_mode = 0 ]
    then
      exit 42
    else
      EXIT_VALUE=42
    fi
  else
    echo "[ OK ]"
  fi
}

function test_case {
  echo "------------------------------------------------------------------------------"
  echo "ENTERING INTO `basename $1` TESTCASE"
  TEST_DIR=$SCRIPT_DIR/`basename $1`
  check_directory "$TEST_DIR"
  check_readable "$TEST_DIR/desc.xml"
  check_readable "$TEST_DIR/nupyf.cmd"
  for ACL_DIR in `ls -d $TEST_DIR/*/`
  do
    test_acl $ACL_DIR
  done
}

delete_output=0
continue_mode=1
test_ldap=1
while getopts "cdhl" flag
do
  #echo "flag: $flag $OPTARG"
  case "$flag" in
    'h') print_usage; exit 0;;
    'l') test_ldap=0;;
    'd') delete_output=1;;
    'c') continue_mode=0;;
    '?') echo "Bad option: $flag"; print_usage; exit 1;;
  esac
done
shift $(($OPTIND-1))

if [ $# = 0 ]
then
  testcases=`cd $SCRIPT_DIR && ls -d */`
else
  testcases=$*
fi

for i in $testcases
do
  test_case $i
done
echo "------------------------------------------------------------------------------"

exit $EXIT_VALUE
